/**
 * Hello, if you have time please have a look at these lines of code.
 * This should be the core code lines of the tutorial, until now with
 * no special features like interactivity or command line output. 
 *
 * Both federates (controlFed, processFed) are very similar. That's why
 * I commented only this file very extensivly. 
 *
 * For now, I think the main purpose should be that the application code is 
 * as easy as possible.
 *
 * Tell me something about your code conventions? Tabs or no tabs etc. .
 *
 *
 *
 */

#include <RTI.hh>
#include <NullFederateAmbassador.hh>
#include <fedtime.hh>

#include <iostream>

// instead
//
// using namespace std;
//
// ?? exact specification ??
using std::string;
using std::cout;
using std::cerr;
using std::endl;
using std::auto_ptr;


// This class realizes the implementation of required fedamb-services
// and holds also the data of incoming callbacks (y).
// Flags are defined additionally. So I can determine the occurence of special
// callbacks in my main().
//
// One such class for both federates is also possible. What's your opinion?
// This would be more straightforward regarding the fedFile.
//
class ControlFedAmb : public NullFederateAmbassador {

  private:


  // another way is to define the variables outside the class as globals
  // so the implementation becomes a little bit shorter. What would you prefer 
  // for the tutorial?
  double y;

  
  bool discovY;
  bool newY;

  public:


  // Here the calling of the parent constructor is not required, what do you think?
  ControlFedAmb() : NullFederateAmbassador() {
    y          = 0;
    discovY    = false;
    newY       = false;
  }

  // these methods are not required if the variables are global
  /** getter */
  double
  getY() {
    return y;
  }

  bool
  getDiscoverY() {
    return discovY;
  }

  bool
  getNewY() {
    return newY;
  }

  /** setter */
  void
  setDiscoverY (bool val) {
    discovY = val;
  }

  void
  setNewY (bool val) {
    newY = val;
  }

  /** required federate services */
  void 
  discoverObjectInstance ( RTI::ObjectHandle theObject,
                           RTI::ObjectClassHandle theObjectClass,
                           const char *theObjectName)
             throw ( RTI::CouldNotDiscover,
                     RTI::ObjectClassNotKnown,
                     RTI::FederateInternalError) {
    cout << "discoverObjInst " << theObjectName << endl;
    if (!strcmp(theObjectName,"y")) {
      discovY = true;
    }
  }

  void 
  reflectAttributeValues( RTI::ObjectHandle, 
                          const RTI::AttributeHandleValuePairSet& 
                                theAttributes,
                          const RTI::FedTime& theTime,
                          const char *theTag, 
                          RTI::EventRetractionHandle theHandle)
            throw ( RTI::ObjectNotKnown, 
                    RTI::AttributeNotKnown, 
                    RTI::FederateOwnsAttributes, 
                    RTI::InvalidFederationTime,
                    RTI::FederateInternalError) {
    RTI::ULong length;
    char *attrValue;

    // I think the iteration through the attributes make no sense here
    // because we have only one. The alternative way is presented 
    // below the for loop.
    for (unsigned int i=0; i<theAttributes.size(); i++) {

      RTI::AttributeHandle handle = theAttributes.getHandle(i);
      length = theAttributes.getValueLength(i);
      attrValue = new char[length];
      theAttributes.getValue(i, attrValue, length);

      if (!strcmp(theTag,"y")) {
        y = *((double*)attrValue);
        newY = true;
      }

      delete[] attrValue;

    }

    //another method
    //double t = ((RTIfedTime&)theTime).getTime();
    //theAttributes.getValue(0, (char*)&y, length);
    //newY = true;
  }
};


// What do you think about exception handling within the
// tutorial? Should every call of an rtiamb-service
// surrounded by a try-catch-construct?

int main() {
  RTI::RTIambassador   rtiAmb; 
  ControlFedAmb        myFedAmb; 

  // hard coding or parameters by command line input?
  string federationName = "TwoLevelController";
  string federateName   = "controlFed";
  string fedFile        = "TwoLevelController.fed";

  /** Federation Management */

  cout << "Create Federation execution " << federationName << "." << endl;

  /** create federation execution */
  try {
    rtiAmb.createFederationExecution(federationName.c_str(), fedFile.c_str());
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }

  cout << "Federate " << federateName << " joins federation execution "
       << federationName << "." << endl;

  /** join federation execution */
  // "spinning while joining" ?? 
  try {
    rtiAmb.joinFederationExecution(federateName.c_str(),
                                   federationName.c_str(),
                                   &myFedAmb);
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }

  /** Declaration Management */

  /** get handles */
  cout << "Get handles from fedFile " << fedFile.c_str() << "." << endl;

  RTI::ObjectClassHandle twoLevelContrID;
  RTI::AttributeHandle uID, yID;

  try {
    twoLevelContrID = rtiAmb.getObjectClassHandle("TwoLevelControllerClass");
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }

  try {
    uID = rtiAmb.getAttributeHandle("U", twoLevelContrID);
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }

  try {
    yID = rtiAmb.getAttributeHandle("Y", twoLevelContrID);
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }

  auto_ptr<RTI::AttributeHandleSet> 
                attrU(RTI::AttributeHandleSetFactory::create(1));
  auto_ptr<RTI::AttributeHandleSet> 
                attrY(RTI::AttributeHandleSetFactory::create(1));

  attrU->add(uID);
  attrY->add(yID);

  /** publish u */
  try {
    rtiAmb.publishObjectClass(twoLevelContrID, *attrU);
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }

  /** subscribe to y */
  try {
    rtiAmb.subscribeObjectClassAttributes(twoLevelContrID, *attrY);
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }

  /** Object Management */

  /** register object */
  RTI::ObjectHandle objInstID_u;

  try {
    objInstID_u = rtiAmb.registerObjectInstance(twoLevelContrID, "u");
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }

  cout << "Wait for second federate." << endl;
  while (!myFedAmb.getDiscoverY()) {
    
    try {
      rtiAmb.tick();
    }
    catch ( RTI::Exception &e ) {
      cerr << "RTI exception: " << e._name << " ["
           << (e._reason ? e._reason : "undefined") << "]." << endl;
    }
    catch ( ... )  {
      cerr << "Error: unknown non-RTI exception." << endl;
    }

    sleep(1);
  }
  myFedAmb.setDiscoverY(false);

  /** simulation loop */

  const double tEnd     = 30;
  const double tStart   = 0;
  const double h        = 0.1;
  const double w        = 0.5;

  double tSim      = tStart;
  double u         = 1;
  double y         = 0;

  auto_ptr<RTI::AttributeHandleValuePairSet>
                ahvps(RTI::AttributeSetFactory::create(1));

  while ( tSim-tEnd <= 0 ) {
    
    y = myFedAmb.getY();

    if (y < w-0.1 ) u=1;
    if (y > w+0.1 ) u=0;

    /** update attribute */
    ahvps -> add(uID, (char*)&(u), sizeof(double));
  
    try {
      rtiAmb.updateAttributeValues(objInstID_u, *ahvps, "u");
    }
    catch ( RTI::Exception &e ) {
      cerr << "RTI exception: " << e._name << " ["
           << (e._reason ? e._reason : "undefined") << "]." << endl;
    }
    catch ( ... )  {
      cerr << "Error: unknown non-RTI exception." << endl;
    }

    ahvps -> empty ();

    /** tick */
    while (!myFedAmb.getNewY()) {

      try {
        rtiAmb.tick();
      }
      catch ( RTI::Exception &e ) {
        cerr << "RTI exception: " << e._name << " ["
             << (e._reason ? e._reason : "undefined") << "]." << endl;
      }
      catch ( ... )  {
        cerr << "Error: unknown non-RTI exception." << endl;
      }

    }
    myFedAmb.setNewY(false);

    cout << "tSim: " << tSim << ", u= " << u << ", y= " << y << endl;
    tSim += h;
  }

  /** Declaration Management */

  // these services are not fundamental, shall I use them, shouldn't I ?
  try {
    rtiAmb.unpublishObjectClass(uID);
  }
  catch ( RTI::Exception &e ) {
     cerr << "RTI exception: " << e._name << " ["
          << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
     cerr << "Error: unknown non-RTI exception." << endl;
  }
  
  try {
    rtiAmb.unsubscribeObjectClass(yID);
  }
  catch ( RTI::Exception &e ) {
     cerr << "RTI exception: " << e._name << " ["
          << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
     cerr << "Error: unknown non-RTI exception." << endl;
  }

  /** Federation Management */

  cout << "Federate " << federateName << " resigns federation execution "
       << federationName << "." << endl;

  /** resign federation exection */
  try {
    rtiAmb.resignFederationExecution(
              RTI::DELETE_OBJECTS_AND_RELEASE_ATTRIBUTES);
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }

  cout << "Destroy Federation execution " << federationName << "." << endl;

  /** destroy federation execution */
  try {  
    rtiAmb.destroyFederationExecution(federationName.c_str());
  }
  catch (RTI::FederatesCurrentlyJoined) {
    cout << "Federates currently joined." << endl;
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }
  
  cout << "Successful termination." << endl;

  return 0;
}
