#include <RTI.hh>
#include <NullFederateAmbassador.hh>
#include <fedtime.hh>

#include <iostream>

// using namespace std;
using std::string;
using std::cout;
using std::cerr;
using std::endl;
using std::auto_ptr;

class ProcessFedAmb : public NullFederateAmbassador {

  private:

  double u;

  bool discovU;
  bool newU;

  public:

  ProcessFedAmb() : NullFederateAmbassador() {
    u          = 0;
    discovU    = false;
    newU       = false;
  }

  /** getter */
  double
  getU() {
    return u;
  }

  bool
  getDiscoverU() {
    return discovU;
  }

  bool
  getNewU() {
    return newU;
  }

  /** setter */
  void
  setDiscoverU (bool val) {
    discovU = val;
  }

  void
  setNewU (bool val) {
    newU = val;
  }

  /** required federate services */
  void 
  discoverObjectInstance ( RTI::ObjectHandle theObject,
                           RTI::ObjectClassHandle theObjectClass,
                           const char *theObjectName)
             throw ( RTI::CouldNotDiscover,
                     RTI::ObjectClassNotKnown,
                     RTI::FederateInternalError) {
    cout << "discoverObjInst " << theObjectName << endl;
    if (!strcmp(theObjectName,"u")) {
      discovU = true;
    }
  }

  void 
  reflectAttributeValues( RTI::ObjectHandle, 
                          const RTI::AttributeHandleValuePairSet& 
                                theAttributes,
                          const RTI::FedTime& theTime,
                          const char *theTag, 
                          RTI::EventRetractionHandle theHandle)
            throw ( RTI::ObjectNotKnown, 
                    RTI::AttributeNotKnown, 
                    RTI::FederateOwnsAttributes, 
                    RTI::InvalidFederationTime,
                    RTI::FederateInternalError) {
    RTI::ULong length;
    char *attrValue;

    for (unsigned int i=0; i<theAttributes.size(); i++) {

      RTI::AttributeHandle handle = theAttributes.getHandle(i);
      length = theAttributes.getValueLength(i);
      attrValue = new char[length];
      theAttributes.getValue(i, attrValue, length);

      if (!strcmp(theTag,"u")) {
        u = *((double*)attrValue);
        newU = true;
      }

      delete[] attrValue;

    }

    //another method
    //double t = ((RTIfedTime&)theTime).getTime();
    //theAttributes.getValue(0, (char*)&u, length);
    //newU = true;
  }
};

int main() {
  RTI::RTIambassador   rtiAmb; 
  ProcessFedAmb        myFedAmb; 

  string federationName = "TwoLevelController";
  string federateName   = "processFed";
  string fedFile        = "TwoLevelController.fed";

  /** Federation Management */

  cout << "Create Federation execution " << federationName << "." << endl;

  /** create federation execution */
  try {
    rtiAmb.createFederationExecution(federationName.c_str(), fedFile.c_str());
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }

  cout << "Federate " << federateName << " joins federation execution "
       << federationName << "." << endl;

  /** join federation execution */
  try {
    rtiAmb.joinFederationExecution(federateName.c_str(),
                                   federationName.c_str(),
                                   &myFedAmb);
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }

  /** Declaration Management */

  /** get handles */
  cout << "Get handles from fedFile " << fedFile.c_str() << "." << endl;

  RTI::ObjectClassHandle twoLevelContrID;
  RTI::AttributeHandle uID, yID;

  try {
    twoLevelContrID = rtiAmb.getObjectClassHandle("TwoLevelControllerClass");
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }

  try {
    uID = rtiAmb.getAttributeHandle("U", twoLevelContrID);
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }

  try {
    yID = rtiAmb.getAttributeHandle("Y", twoLevelContrID);
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }

  auto_ptr<RTI::AttributeHandleSet> 
                attrU(RTI::AttributeHandleSetFactory::create(1));
  auto_ptr<RTI::AttributeHandleSet> 
                attrY(RTI::AttributeHandleSetFactory::create(1));

  attrU->add(uID);
  attrY->add(yID);

  /** publish y */
  try {
    rtiAmb.publishObjectClass(twoLevelContrID, *attrY);
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }

  /** subscribe to u */
  try {
    rtiAmb.subscribeObjectClassAttributes(twoLevelContrID, *attrU);
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }

  /** Object Management */

  /** register object */
  RTI::ObjectHandle objInstID_y;

  try {
    objInstID_y = rtiAmb.registerObjectInstance(twoLevelContrID, "y");
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }

  cout << "Wait for second federate." << endl;
  while (!myFedAmb.getDiscoverU()) {
    
    try {
      rtiAmb.tick();
    }
    catch ( RTI::Exception &e ) {
      cerr << "RTI exception: " << e._name << " ["
           << (e._reason ? e._reason : "undefined") << "]." << endl;
    }
    catch ( ... )  {
      cerr << "Error: unknown non-RTI exception." << endl;
    }

    sleep(1);
  }
  myFedAmb.setDiscoverU(false);

  /** simulation loop */

  const double tEnd     = 30;
  const double tStart   = 0;
  const double h        = 0.1;
  const double w        = 0.5;
  const double K        = 1;
  const double T1       = 5;

  double tSim      = tStart;
  double u         = 1;
  double y         = 0;
  double ydot      = 0;

  auto_ptr<RTI::AttributeHandleValuePairSet>
                ahvps(RTI::AttributeSetFactory::create(1));

  while ( tSim-tEnd <= 0 ) {

    while (!myFedAmb.getNewU()) {
      try {
        rtiAmb.tick();
      }
      catch ( RTI::Exception &e ) {
        cerr << "RTI exception: " << e._name << " ["
             << (e._reason ? e._reason : "undefined") << "]." << endl;
      }
      catch ( ... )  {
        cerr << "Error: unknown non-RTI exception." << endl;
      }
    }
    myFedAmb.setNewU(false);

    u = myFedAmb.getU();
			    
    /* compute y by explicit Euler */
    ydot = ( -y + u*K ) / T1;
    y = y + h*ydot;

    /** update attribute */
    ahvps -> add(yID, (char*)&(y), sizeof(double));
  
    try {
      rtiAmb.updateAttributeValues(objInstID_y, *ahvps, "y");
    }
    catch ( RTI::Exception &e ) {
      cerr << "RTI exception: " << e._name << " ["
           << (e._reason ? e._reason : "undefined") << "]." << endl;
    }
    catch ( ... )  {
      cerr << "Error: unknown non-RTI exception." << endl;
    }

    ahvps -> empty ();

    cout << "tSim: " << tSim << ", u= " << u << ", y= " << y << endl;
    tSim += h;
  }

  /** Declaration Management */
  try {
    rtiAmb.unpublishObjectClass(yID);
  }
  catch ( RTI::Exception &e ) {
     cerr << "RTI exception: " << e._name << " ["
          << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
     cerr << "Error: unknown non-RTI exception." << endl;
  }
  
  try {
    rtiAmb.unsubscribeObjectClass(uID);
  }
  catch ( RTI::Exception &e ) {
     cerr << "RTI exception: " << e._name << " ["
          << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
     cerr << "Error: unknown non-RTI exception." << endl;
  }

  /** Federation Management */

  cout << "Federate " << federateName << " resigns federation execution "
       << federationName << "." << endl;

  /** resign federation exection */
  try {
    rtiAmb.resignFederationExecution(
              RTI::DELETE_OBJECTS_AND_RELEASE_ATTRIBUTES);
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }

  cout << "Destroy Federation execution " << federationName << "." << endl;

  /** destroy federation execution */
  try {  
    rtiAmb.destroyFederationExecution(federationName.c_str());
  }
  catch (RTI::FederatesCurrentlyJoined) {
    cout << "Federates currently joined." << endl;
  }
  catch ( RTI::Exception &e ) {
    cerr << "RTI exception: " << e._name << " ["
         << (e._reason ? e._reason : "undefined") << "]." << endl;
  }
  catch ( ... )  {
    cerr << "Error: unknown non-RTI exception." << endl;
  }
  
  cout << "Successful termination." << endl;

  return 0;
}
